/*
 * (C) Copyright 2001 by Thomas Driemeyer <thomas@bitrot.de> www.bitrot.de.
 *
 * This is free software; you can redistribute it and/or modify it under the
 * terms of the GNU Library General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
 * General Public License for more details.
 *
 *	Config::Config
 *	Config::read
 *	Config::write
 *	Config::get_history
 *	Config::set_history
 *	Config::insert_history
 *	Config::get_lasthistory
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "config.h"

static int qpebug;	// QPE apps go through a phase of insanity at
			// startup. Ignore impossible method calls then.


//
// initialize empty Config
//

Config::Config()
{
	modified = false;
	lasthistory = -1;
	memset(history, 0, sizeof(history));
	qpebug = 1;
}


//
// read the config file from disk. If the file doesn't exist, use defaults,
// that's not an error. Make sure that config_dir exists and cd to it. If
// you think that qpebug is seriously nuts, you are right - in QPE mode,
// the QPEApplication constructor somehow manages to call Config::read
// before Config::Config! Multiple times! This crashes the program if a
// config file exists. The QApplication constructor is sane, so I can't
// debug this. Annoying.
//

bool Config::read()
{
	FILE		*fp;		// open file
	char		line[1100];	// line buffer
	char		*seed;		// for finding seed in history line
	char		*home;		// for cd to home dir
	int		hi = 0;		// history index
	int		i;		// for removing trailing whitespace

	if (qpebug != 1)				// constructor done?
		return(false);
	qpebug = 2;					// read only once...
	if (home = getenv("HOME"))			// go home
		chdir(home);
	if (*config_dir) {
		mkdir(config_dir, 0700);		// create config dir
		chdir(config_dir);			// enter config dir
	}
	if (!(fp = fopen(config_file, "r")))
		return(true);
	while (fgets(line, sizeof(line), fp)) {		// line loop
		for (i=strlen(line)-1; i >= 0; i--)
			if (line[i]==' ' || line[i]=='\t' || line[i]=='\n')
				line[i] = 0;
			else
				break;

		if (!strncmp(line, "last=", 5))
			lasthistory = atoi(line+5);

		else if (!strncmp(line, "history=md", 10) && hi < nhistory) {
			History *hist = &history[hi++];
			seed = strchr(line+12, ',');
			seed = seed ? seed+1 : line+12;
			hist->md4     = line[10] == '4';
			hist->number  = atoi(line+12);
			hist->seed    = new char[strlen(seed)+1];
			hist->phrase  = 0;
			strcpy(hist->seed, seed);
		}
	}
	fclose(fp);
	if (lasthistory < 0 || lasthistory >= nhistory)
		lasthistory = hi-1;
	return(true);
}


//
// write the config file to disk so we can start up the next time as we left.
// It's tempting to also save the passphrase here but DON'T DO THAT. The whole
// point of OTP passwords is that they are more secure than regular passwords,
// and if security matters you absolutely don't want to save passwords to a
// file! Encrypting wouldn't help much because if read() can decrypt it, so
// can an intruder.
//

bool Config::write()
{
	FILE		*fp;		// open file
	bool		ok = true;	// writing ok so far?

	if (!modified)
		return(true);
	if (!(fp = fopen(config_file, "w"))) {
		perror(config_file);
		return(false);
	}
	ok &= fprintf(fp, "last=%d\n", lasthistory);
	for (int hi=0; hi < nhistory && history[hi].seed; hi++) {
		History *hist = &history[hi];
		ok &= fprintf(fp, "history=md%d,%d,%s\n",
			hist->md4 ? 4 : 5, hist->number, hist->seed);
	}
	ok &= !fclose(fp);
	modified = !ok;
	return(ok);
}


//
// return a history entry, or set a history entry. When replacing an entry,
// it can happen that the new and the old entry are identical as the result
// of choosing an entry from the seed pulldown; in this case the seed argument
// pointer and the history seed pointer are identical. See slot_seed().
//

const History *Config::get_history(
	int		hi)		// get this history entry
{
	return(hi >= 0 && hi < nhistory && history[hi].seed ? &history[hi] :0);
}


void Config::set_history(
	int		hi,		// set this history entry
	bool		md4,		// use MD4 instead of MD5
	int		number,		// new number value
	const char	*seed,		// new seed value
	const char	*phrase)	// new phrase value
{
	if (hi < 0 || hi >= nhistory)
		return;
	History *hist = &history[hi];
	if (hist->seed && hist->seed != seed)
		delete[] hist->seed;
	if (hist->phrase)
		delete[] hist->phrase;
	if (seed && phrase) {
		hist->md4    = md4;
		hist->number = number;
		hist->seed   = new char[strlen(seed)+1];
		hist->phrase = new char[strlen(phrase)+1];
		strcpy(hist->seed,   seed);
		strcpy(hist->phrase, phrase);
	} else {
		hist->md4    = false;
		hist->number = 0;
		hist->seed   = 0;
		hist->phrase = 0;
	}
	lasthistory = hi;
	modified = true;
}


//
// insert history entry: find the seed in the list. If it's there, update
// the entry. Otherwise, append it to the end; if the list is full drop the
// first entry.
//

void Config::insert_history(
	bool		md4,		// use MD4 instead of MD5
	int		number,		// new number value
	const char	*seed,		// new seed value
	const char	*phrase)	// new phrase value
{
	int		hi;		// history scan index

	for (hi=0; hi < nhistory; hi++)
		if (!history[hi].seed || !strcmp(history[hi].seed, seed))
			break;
	if (hi == nhistory) {
		set_history(0, false, 0, 0, 0);
		for (hi=0; hi < nhistory-1; hi++)
			history[hi] = history[hi+1];
		history[hi].seed = history[hi].phrase = 0;
	}
	set_history(hi, md4, number, seed, phrase);
}


//
// return the last history entry used. This is useful after startup, to put
// the last used number and seed into the GUI.
//

int Config::get_lasthistory()
{
	return(lasthistory);
}
